// projet1.c
//
// HEROUARD Brice (brice.herouard@utbm.fr)
// COUVIDOU Laurent (laurent.couvidou@utbm.fr)
//
// RSOLUTION D'UN SYSTEME D'QUATIONS PAR LA METHODE DU PIVOT MAXIMUM
// Systme d'quations AX=B, n quations, p inconnues
// A : matrice des coefficients (n lignes, p colonnes)
// B : matrice des constantes (n lignes)
// X : matrice des solutions (p colonnes)

#include <stdio.h> // Librairie C standard

#define TAI 20 // Taille des matrices


/******************** DECLARATIONS DE FONCTIONS *******************************/


/****** Affichage *************************************************************/

void Affichage1 ( float M[TAI], int a );
// Affichage d'une matrice ligne (ou colonne) de a lments

void Affichage2 ( float M[TAI][TAI], int a, int b );
// Affichage d'une matrice de a lignes et b colonnes

void AffichageInc ( int pri[TAI], int* compteur_pri, int aux[TAI], int* compteur_aux );
// Dans le cas d'un systme indetermin (cas=2), affiche les inconnues
// principales et auxiliaires

void AffichageInd ( float MAT[TAI][TAI], int p );
// Dans le cas d'un systme indtermin, affiche un systme simplifi,
// prsentant les inconnues principales en fonction des inconnues auxiliaires


/****** Saisie ****************************************************************/

void SaisieAB ( float A[TAI][TAI], float B[TAI], int n, int p );
// Saisie des matrices A et B


/****** Test ******************************************************************/

int Tester(float A[TAI][TAI], float B[TAI],int n, int p);
// Recherche les quations du type 0x1 + 0x2 + ... + 0xp = constante
// Renvoie un entier :
//   1 -> Il en existe, le systme ne peut pas tre rsolu
//   0 -> Le systme peut tre rsolu


/****** Simplification ********************************************************/

int SimplifNulles(float A[TAI][TAI], float B[TAI], int n, int p);
// Recherche les quations du type 0x1 + 0x2 + ... + 0xp = 0
// Supprime ces quations du systme
// Renvoie n, nombre d'quations restantes

int SimplifEgales(float A[TAI][TAI], float B[TAI], int n, int p);
// Recherche les quations gales  un coeffcient prs
// Supprime ces quations du systme
// Renvoie n

int SupprimLigne(float A[TAI][TAI], float B[TAI], int n, int p, int l);
// Supprime la ligne l du systme, retranche 1  n
// Renvoie n


/****** Pivot maximum *********************************************************/

int MaxLigne ( float A[TAI][TAI], int p, int l );
// Recherche le coefficient maximum d'une ligne de A
// Le renvoie

void DivLigne ( float A[TAI][TAI], float B[TAI], int p, int l, float div );
// Divise une ligne de A et de B par div

void SimplifLigne (  float A[TAI][TAI], float B[TAI], int p, int lsimp, int lbase , int c );
// Coeur du "pivot maximum" :
// Retranche la ligne lbase, fois le c-ime coefficient de lsimp,  la ligne lsimp


/****** Rsolution ************************************************************/

float Somme ( float A[TAI][TAI], float X[TAI], int p, int l );
// Ralise la somme de la l-ime ligne de A, except le c-ime coefficient

void Resoudre1 ( float A[TAI][TAI], float B[TAI], float X[TAI], int n, int p, int c[TAI]);
// Rsoud un systme  solutions uniques (n=p), place les solutions dans X

void Resoudre2 ( float A[TAI][TAI], float B[TAI],
				 int pri[TAI], int* compteur_pri, int aux[TAI], int* compteur_aux,
				 float MAT[TAI][TAI],
				 int n, int p, int c[TAI]);
// Rsoud un systme indtermin (n<p), place les inconnues principales dans
// pri et les inconnues auxiliaires dans aux.

/****** Divers ****************************************************************/

float ValAbs ( float v );
// Renvoie la valeur absolue du rel v
// vite l'inclusion d'une bibliothque complte pour une seule fonction


/******************** PROGRAMME PRINCIPAL *************************************/


int main(int argc, char** argv){

	float A[TAI][TAI], B[TAI];
	int n, p;
	// Donnes du problme

	float X[TAI];
	int pri[TAI];		// Inconnues principales
	int compteur_pri=0;	// Leur nombre
	int aux[TAI];		// Inconnues auxiliaires
	int compteur_aux=0;	// Leur nombre
	float MAT[TAI][TAI];
	// Rsultats du problme

	int i, j;
	// Compteurs

	int cas;
	// Donne le "type" du systme :
    //   -> cas=1 : le systme a une solution unique
	//   -> cas=2 : Le systme est indetermin
	//   -> cas=3 : Le systme est impossible

	int indmax[TAI];
	// Donne, pour chaque ligne, l'indice de la case o l'on met un 1 lors de
	// l'utilisation du pivot maximum (cf ci-dessous)

	printf("*** Resolution d un systeme d equations par la methode du pivot maximum ***\n\n");

	/* Acquisition des donnes */

	printf("Entrez n, nombre d'equations (entier) : ");
	scanf("%d",&n);
	printf("Entrez p, nombre d'inconnues (entier) : ");
	scanf("%d",&p);

    printf("\n");

	SaisieAB ( A, B, n, p );

	printf("\nA, matrice des coefficients\n");
	Affichage2 ( A, n, p );

	printf("\nB, matrice des constantes\n");
	Affichage1 ( B, n );

	printf("\n");

	/* Test du cas */

	if (Tester(A, B, n, p)){
         cas = 3;
    }

    else{

     	n = SimplifNulles ( A, B, n, p);
	    n = SimplifEgales ( A, B, n, p);

		if (n==p){
		    cas=1;
        }
        else{
            cas=2;
        }

		/* Pivot maximum */

		int i, j;

 	    for(i=0;i<n;i++){

		 	indmax[i] = MaxLigne ( A, p, i );

			DivLigne ( A, B, p, i, A[i][indmax[i]] );

			for(j=i+1;j<n;j++){

			   SimplifLigne ( A, B, p, j, i, indmax[i] );

			}
		}
	}

	/* Rsolution et affichage du rsultat */

	switch(cas){

        case 1 :
				for(i=0;i<TAI;i++){
					X[i]=0;
    			}
    			// Place des 0 dans X

                printf("Le systeme a pour solutions :\n\n");

				Resoudre1 ( A, B, X, n, p, indmax );
                Affichage1 ( X, n );

                break;

        case 2 :
                printf("Le systeme est indetermine.\n");
				Resoudre2 ( A, B, pri, &compteur_pri, aux, &compteur_aux, MAT, n, p, indmax );
                AffichageInc ( pri, &compteur_pri, aux, &compteur_aux );
				AffichageInd ( MAT, p );

                break;

        case 3 :
                printf("Le systeme n'a pas de solution.\n");

                break;

        default :
                ;

    }

	system("PAUSE");
	// Pause avant de mettre fin au programme

	return 0;
	// Code de retour du programme : aucune erreur

}


/******************** DEFINITIONS DE FONCTIONS ********************************/


/****** Affichage *************************************************************/

void Affichage1 ( float M[TAI], int a ) {

	int i;

	for(i=0;i<a;i++){
		printf("%+6.3f\n",M[i]);
	}
}

void Affichage2 ( float M[TAI][TAI], int a, int b ) {

	int i, j;

	for(i=0;i<a;i++){
		for(j=0;j<b;j++){
			printf("%+6.3f ",M[i][j]);
		}
		printf("\n");
	}
}

void AffichageInc ( int pri[TAI], int *compteur_pri, int aux[TAI], int *compteur_aux ){

	int i;

	printf("\nLes inconnue(s) principale(s) sont : ");
	i=0;
	while(i<*compteur_pri){
		printf(" x%d ",pri[i]);
		i++;
	}

	printf("\nLes inconnue(s) auxiliaire(s) sont : ");
	i=0;
	while(i<*compteur_aux){
		printf(" x%d ",aux[i]);
		i++;
	}

	printf("\n");

}

void AffichageInd ( float MAT[TAI][TAI], int p ){

	int i=0, j, k;
	float coeff;
	float TMP[TAI][TAI];

	while(i<p){

		k=0;

		while(k<p){

			j=0;
			coeff=MAT[i][k];

			while(j<p){

				TMP[k][j]=MAT[k][j]*coeff;
				if (k!=i) TMP[k][p]=MAT[k][p]*coeff;
				else TMP[k][p]=MAT[k][p];
				j++;
			}

			k++;
		}

		if (MAT[i][p+1]==1){

			printf("x%i = ",i);

			int a=0, compteur=0;

			while(a<=p){

   			   int b=0;
               float somme=0;

			   while(b<p){
                   somme=somme+TMP[b][a];
                   b++;
               }

               if (somme!=0){
               		compteur=1;
					printf(" %+6.3f",somme);
					if (a<p){
					     printf("x%i ",a);
					}
			   }

			   if ((a==p)&&(compteur==0)){
			         printf("0 ");
			   }

			   a++;
			}

			printf("\n");

		}

		i++;
	}
}


/****** Saisie ****************************************************************/

void SaisieAB ( float A[TAI][TAI], float B[TAI], int n, int p ) {

	int i, j;

	printf("Saisie de A, matrice des coefficients\n\n");
	for(i=0;i<n;i++){
		for(j=0;j<p;j++){
			printf("a%d%d = ",i,j);
			scanf("%f",&A[i][j]);
		}
		printf("\n");
	}

	printf("Saisie de B, matrice des constantes\n\n");
	for(i=0;i<n;i++){
		printf("b%d = ",i);
		scanf("%f",&B[i]);
	}

}


/****** Test ******************************************************************/

int Tester(float A[TAI][TAI], float B[TAI],int n, int p){

	int h=0, i=0, retour=0, test=0;

    while((i<n)&&(n>=0)){

	    int compteur=0, continuer=1;

        if (B[i]!=0){
            compteur++;
        }
        else{
            continuer=0;
        }

        int j=0;

        while((j<p)&&(continuer)){

        	if (A[i][j]!=0){
               	continuer=0;
       		}
           	else{
               	compteur++;
           	}

           	j++;
    	}

        if (compteur==p+1){
            test=1;
        }

        i++;
    }

	if (test==1){
 	    return 1;
 	}
    else{
        return 0;
    }
}


/****** Simplification ********************************************************/

int SimplifNulles(float A[TAI][TAI], float B[TAI], int n, int p){

    int i=0;

	while((i<n)&&(n>=0)){

	    int j=0, compteur=0, stop=0;

		if (B[i]!=0){
		    stop=1;
        }
        else{
            compteur++;
        }

		while((!stop)&&(j<p)){

			if (A[i][j]!=0){
			    stop=1;
            }
			else{
                compteur++;
            }
            j++;
        }

		if ((!stop)&&(compteur==p+1)){

			n=SupprimLigne ( A, B, n, p, i);
			i=-1;
		}

		i++;
	}

	return(n);
}

int SimplifEgales(float A[TAI][TAI],float B[TAI], int n, int p){

	int h=0;

	while(h<n-1){

	    int i=h+1;
	    float coeffA, coeffB;

		while(i<n){

			int compteur=0, stop=0, j=0;

			if ((B[i]==0)&&(B[h]!=0)){
			    stop=1;
			}
			else if ((B[i]==0)&&(B[h]==0)){
				coeffB=0;
			}
			else{
                coeffB=B[h]/B[i];
            }

			while((j<p)&&(!stop)){

				if ((A[i][j]==0)&&(A[h][j]==0)){
					coeffA=0;
				}
				else if ((A[i][j]==0)&&(A[i][j]!=0)){
					stop=1;
				}
			    else{
                    coeffA=(A[h][j]/A[i][j]);
                }

				if (coeffA==coeffB){
                    compteur++;
				}

				j++;
			}

			if (compteur==p){
				n=SupprimLigne( A, B, n, p, i );
				i=h;
			}

			i++;
		}

		h++;
	}

	return(n);
}

int SupprimLigne(float A[TAI][TAI], float B[TAI], int n, int p, int l){

	while(l<n-1){

	    B[l]=B[l+1];

		int c=0;
		while(c<p){
			A[l][c]=A[l+1][c];
			c++;
		}

		l++;
	}

	return n-1;

}

/****** Pivot maximum *********************************************************/

int MaxLigne ( float A[TAI][TAI], int p, int l ) {

	int i, indice=0;

	float maxabs=ValAbs(A[l][0]);

	for(i=1;i<p;i++){
		if(ValAbs(A[l][i])>maxabs){
			maxabs=ValAbs(A[l][i]);
			indice=i;
		}
	}

	return indice;

}

void DivLigne ( float A[TAI][TAI], float B[TAI], int p, int l, float div ) {

	int i;

	for (i=0;i<p;i++) {

		A[l][i]=A[l][i]/div;

	}

	B[l]=B[l]/div;

}

void SimplifLigne (  float A[TAI][TAI], float B[TAI], int p, int lsimp, int lbase , int c ) {

	int i;
	float coeff=A[lsimp][c];

	for (i=0;i<p;i++) {

		A[lsimp][i]=A[lsimp][i]-coeff*A[lbase][i];

	}

	B[lsimp]=B[lsimp]-coeff*B[lbase];

}


/****** Rsolution ************************************************************/

float Somme ( float A[TAI][TAI], float X[TAI], int p, int l ) {

	int i;
	float somme=0;

	for(i=0;i<p;i++){
		somme = somme + A[l][i] * X[i];
	}

	return somme;

}

void Resoudre1 ( float A[TAI][TAI], float B[TAI], float X[TAI], int n, int p, int c[TAI] ) {

	int i;

	for(i=n-1;i>=0;i--){

		X[c[i]] = ( B[i] - Somme( A, X, p, i ) );

	}

}

void Resoudre2 ( float A[TAI][TAI], float B[TAI],
				 int pri[TAI], int* compteur_pri, int aux[TAI], int* compteur_aux,
				 float MAT[TAI][TAI],
				 int n, int p, int c[TAI]) {

	int i,j,k,test;

	int compteur;

	for(i=0;i<TAI;i++){
		for(j=0;j<TAI;j++){
			MAT[i][j]=0;
		}
    }

	i=n-1;

	while(i>=0){

		compteur=0;
		j=0;
		while(j<p){
			if ((A[i][j]==1)&&(compteur==0)){
				compteur=1;
				pri[i]=j;
				MAT[j][p+1]=1;
				(*compteur_pri)++;
			}
			else{
			    j++;
   			}
		}
		i--;
	}

	i=0;

	while(i<p){

		if (MAT[i][p+1]==0){
			aux[*compteur_aux]=i;
			(*compteur_aux)++;
		}
		i++;
	}

	i=n-1;

	while(i>=0){

		j=0;
		test=0;
		while(j<p){
			if ((A[i][j]==1)&&(test==0)){
				test=1;
				MAT[j][p]=B[i];
				k=0;
				while(k<p){
					if (j!=k) MAT[j][k]=-A[i][k];
					else MAT[j][k]=0;
					k++;
				}
			}
			j++;
		}
		i--;
	}

	i=0;

	while(i<*compteur_aux){
		MAT[aux[i]][aux[i]]=1;
		i++;
	}
}


/****** Divers ****************************************************************/

float ValAbs ( float v ) {
	if ( v<0 ) return -v;
	else return v;
}
